from __future__ import annotations

import math
from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd

from krisi.utils.data import generate_synthetic_predictions_binary, shuffle_df_in_chunks

from ..type import Model, WeightsDS


class RandomClassifier(Model):
    def __init__(self) -> None:
        self.name = "NS"

    def predict(
        self, X: pd.DataFrame, y: pd.Series, sample_weight: Optional[WeightsDS] = None
    ) -> Tuple[pd.Series, pd.DataFrame]:
        preds_probs = generate_synthetic_predictions_binary(y, sample_weight)
        predictions = preds_probs.iloc[:, 0]
        probabilities = preds_probs.iloc[:, 1:3]
        return predictions, probabilities


class RandomClassifierSmoothed(Model):
    def __init__(self, smoothing_window: Optional[int] = None) -> None:
        self.name = "NS-Smooth"
        self.smoothing_window = smoothing_window

    def predict(
        self, X: pd.DataFrame, y: pd.Series, sample_weight: Optional[WeightsDS] = None
    ) -> Tuple[pd.Series, pd.DataFrame]:
        def get_avarage_change_length(ds: pd.Series) -> float:
            # This only works for binary classification
            indices = ds.diff().fillna(0).ne(0).cumsum()
            differences = indices.groupby(indices).size()
            return differences.mean()

        smoothing_window = (
            math.floor(get_avarage_change_length(y))
            if self.smoothing_window is None
            else self.smoothing_window
        )

        preds_probs = generate_synthetic_predictions_binary(
            y, sample_weight, smoothing_window=smoothing_window
        )
        predictions = preds_probs.iloc[:, 0]
        probabilities = preds_probs.iloc[:, 1:3]
        return predictions, probabilities


class RandomClassifierChunked(Model):
    def __init__(self, chunk_size: Union[float, int]) -> None:
        self.name = "NS"
        self.chunk_size = chunk_size

    def predict(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weight: Optional[WeightsDS] = None,
    ) -> Tuple[pd.Series, pd.DataFrame]:
        X = shuffle_df_in_chunks(X, self.chunk_size)
        predictions = X.iloc[:, 0]
        probabilities = X.iloc[:, 1:3]
        return predictions, probabilities


class PerfectModel(Model):
    def __init__(self) -> None:
        self.name = "PM"

    def predict(
        self, X: pd.DataFrame, y: pd.Series, sample_weight: WeightsDS
    ) -> Tuple[pd.Series, pd.DataFrame]:
        predictions = y
        probabilities = pd.get_dummies(y, dtype=float)
        return predictions, probabilities


class WorstModel(Model):
    def __init__(self) -> None:
        self.name = "WM"

    def predict(
        self, X: pd.DataFrame, y: pd.Series, sample_weight: WeightsDS
    ) -> Tuple[pd.Series, pd.DataFrame]:
        predictions = np.logical_xor(y, 1).astype(int)
        probabilities = pd.get_dummies(predictions, dtype=float)
        return predictions, probabilities


class ModelRegistry:
    RandomClassifier = RandomClassifier
    RandomClassifierSmoothed = RandomClassifierSmoothed
    RandomClassifierChunked = RandomClassifierChunked
    PerfectModel = PerfectModel
    WorstModel = WorstModel



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    TypeVar,
    Union,
)

import numpy as np
import pandas as pd

from krisi.evaluate.type import Predictions, Targets

if TYPE_CHECKING:
    from krisi.evaluate.metric import Metric
    from krisi.report.type import InteractiveFigure


def map_newdict_on_olddict(
    old_dict: dict, new_dict: dict, exclude: List[str] = []
) -> dict:
    merged_dict = old_dict.copy()
    for key_, value_ in new_dict.items():
        if key_ in old_dict and value_ is not None and key_ not in exclude:
            merged_dict[key_] = value_

    return merged_dict


def group_by_categories(
    flat_list: List[Union["Metric", "InteractiveFigure"]], categories: List[str]
) -> Dict[str, Union["Metric", "InteractiveFigure"]]:
    category_groups = dict()
    for category in categories:
        category_groups[category] = list(
            filter(
                lambda x: x.category.value == category
                if (hasattr(x, "category") and x.category is not None)
                else False,
                flat_list,
            )
        )
    category_groups[None] = list(
        filter(
            lambda x: x.category is None if hasattr(x, "category") else False,
            flat_list,
        )
    )
    return category_groups


def type_converter(type_to_ensure: Any) -> Callable:
    def ensure_format(
        obj_s: Union[List[Any], Any]
    ) -> Union[List[type_to_ensure], type_to_ensure]:
        if isinstance(obj_s, Iterable) and not isinstance(obj_s, str):
            return [
                type_to_ensure(el) if not isinstance(el, type_to_ensure) else el
                for el in obj_s
            ]
        else:
            if isinstance(obj_s, type_to_ensure):
                return obj_s
            else:
                return type_to_ensure(obj_s)

    return ensure_format


def string_to_id(s: str) -> str:
    s = "".join(filter(lambda c: str.isidentifier(c) or str.isdecimal(c), s))
    return s


def isiterable(obj: Any) -> bool:
    if isinstance(obj, Iterable) and not isinstance(obj, str):
        return True
    else:
        return False


def isnumber(obj: Any) -> bool:
    if isinstance(obj, (int, float, np.number)):
        return True
    else:
        return False


def check_iterable_with_number(iterable: Any):
    if isinstance(iterable, Exception) or iterable is None:
        return False

    if isiterable(iterable) and len(iterable) > 0:
        if isnumber(iterable[0]):
            return True
    return False


def strip_builtin_functions(dict_to_strip: dict) -> dict:
    return {key_: value_ for key_, value_ in dict_to_strip.items() if key_[:2] != "__"}


def __flatten(xs: Iterable[Any]) -> Any:
    for x in xs:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x


def flatten(xs: Iterable[Any]) -> Iterable[Any]:
    return list(__flatten(xs))


def remove_nans(iter):
    if isinstance(iter, List):
        return [el for el in iter if el is not None]
    else:
        return {key: el for key, el in iter.items() if el is not None}


def calculate_nans(ds: Union[Targets, Predictions]) -> int:
    return ds.isna().sum() if isinstance(ds, pd.Series) else sum(np.isnan(ds))


T = TypeVar("T")


def wrap_in_list(input: Union[T, List[T]]) -> List[T]:
    return input if isinstance(input, List) else [input]


def replace_if_none(input: Any, replacement: Any) -> Any:
    return replacement if input is None else input


def is_int(s: Any):
    if isinstance(s, str) and "." in s:
        return False
    try:
        int(s)
    except ValueError:
        return False
    else:
        return True


def del_dict_keys(d: dict, keys: Union[str, List[str]]) -> dict:
    keys = wrap_in_list(keys)
    return {k: v for k, v in d.items() if k not in keys}


def empty_if_none(metrics: Optional[List]) -> List:
    return metrics if metrics is not None else []


def filter_none(list: List) -> List:
    return [el for el in list if el is not None]



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

import numpy as np
import pandas as pd

from krisi.evaluate.type import (
    Model,
    PredictionsDS,
    ProbabilitiesDF,
    Purpose,
    TargetsDS,
    WeightsDS,
)

if TYPE_CHECKING:
    from krisi.evaluate.metric import Metric


def zscore(arr: np.ndarray) -> np.ndarray:
    m = arr.mean()
    s = arr.std(ddof=0)
    return (arr - m) / s


def _predict_evaluate(
    X: pd.DataFrame,
    y: pd.Series,
    sample_weight: Optional[pd.Series],
    metric: Metric,
    model: Model,
) -> float:
    preds, probs = model.predict(X, y, sample_weight)
    return metric.evaluate(y, preds, probs, sample_weight).result


def calculate_benchmark(
    metric: Metric,
    models: List[Model],
    y: TargetsDS,
    predictions: PredictionsDS,
    probabilities: Optional[ProbabilitiesDF],
    sample_weight: Optional[WeightsDS],
    num_benchmark_iter: int,
) -> Metric:
    if metric.purpose == Purpose.group or metric.purpose == Purpose.diagram:
        return metric

    metric_result = (
        metric.evaluate(y, predictions, probabilities, sample_weight).result
        if metric.result is None
        else metric.result
    )
    benchmark_metric = metric.reset()

    for model in models:
        if probabilities is not None:
            X = pd.concat([predictions, probabilities], axis="columns", copy=False)
        else:
            X = predictions.to_frame()

        benchmark_metrics = [
            _predict_evaluate(X, y, sample_weight, benchmark_metric, model)
            for _ in range(num_benchmark_iter)
        ]

        mean_benchmark_metric = np.mean(benchmark_metrics)

        if metric.purpose == Purpose.objective:
            comparison_result = metric_result - mean_benchmark_metric
        elif metric.purpose == Purpose.loss:
            comparison_result = mean_benchmark_metric - metric_result

        benchmark_zscores = zscore(np.array(benchmark_metrics + [metric_result]))
        metric_zscore = benchmark_zscores[-1]

        metric.comparison_result = pd.concat(
            [
                metric.comparison_result,
                pd.Series([comparison_result], index=[f"Δ {model.name}"]),
                pd.Series([metric_zscore], index=[f"Δ {model.name}_zscore"]),
            ],
            axis=0,
        )

    return metric



from copy import deepcopy
from typing import List

import numpy as np
import pandas as pd
from sklearn.metrics import (
    mean_absolute_error,
    mean_absolute_percentage_error,
    mean_squared_error,
    mean_squared_log_error,
    r2_score,
)

from ..group import Group
from ..metric import Metric
from ..type import ComputationalComplexity, MetricCategories, Purpose
from .diagrams import (
    display_acf_plot,
    display_density_plot,
    display_single_value,
    display_time_series,
)

mae = Metric[float](
    name="Mean Absolute Error",
    key="mae",
    category=MetricCategories.reg_err,
    info="(Mean absolute error) represents the difference between the original and predicted values extracted by averaged the absolute difference over the data set.",
    func=lambda y, pred, **kwargs: mean_absolute_error(y, pred, **kwargs),
    plot_funcs=[(display_single_value, dict(width=900.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    purpose=Purpose.loss,
)
""" ~ """

mape = Metric[float](
    name="Mean Absolute Percentage Error",
    key="mape",
    category=MetricCategories.reg_err,
    func=lambda y, pred, **kwargs: mean_absolute_percentage_error(y, pred, **kwargs),
    plot_funcs=[(display_single_value, dict(width=900.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    purpose=Purpose.loss,
)
""" ~ """

smape = Metric[float](
    name="Symmetric Mean Absolute Percentage Error",
    key="smape",
    category=MetricCategories.reg_err,
    func=lambda y_true, y_pred, **kwargs: np.mean(
        np.abs(y_pred - y_true) / (np.abs(y_true) + np.abs(y_pred)), axis=0
    ),
    plot_funcs=[(display_single_value, dict(width=900.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    purpose=Purpose.loss,
)
""" ~ """

mse = Metric[float](
    name="Mean Squared Error",
    key="mse",
    category=MetricCategories.reg_err,
    info="(Mean Squared Error) represents the difference between the original and predicted values extracted by squared the average difference over the data set.",
    parameters={"squared": True},
    func=lambda y, pred, **kwargs: mean_squared_error(y, pred, **kwargs),
    plot_funcs=[(display_single_value, dict(width=900.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    purpose=Purpose.loss,
)
""" ~ """

rmse = Metric[float](
    name="Root Mean Squared Error",
    key="rmse",
    category=MetricCategories.reg_err,
    info="(Root Mean Squared Error) is the error rate by the square root of Mean Squared Error.",
    parameters={"squared": False},
    func=lambda y, pred, **kwargs: mean_squared_error(y, pred, **kwargs),
    plot_funcs=[(display_single_value, dict(width=900.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    purpose=Purpose.loss,
)
""" ~ """

rmsle = Metric[float](
    name="Root Mean Squared Log Error",
    key="rmsle",
    category=MetricCategories.reg_err,
    parameters={"squared": False},
    func=lambda y, pred, **kwargs: mean_squared_log_error(y, pred, **kwargs),
    plot_funcs=[(display_single_value, dict(width=900.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    purpose=Purpose.loss,
)
""" ~ """

r_two = Metric[float](
    name="R-squared",
    key="r_two",
    category=MetricCategories.reg_err,
    info="(Coefficient of determination) represents the coefficient of how well the values fit compared to the original values. The value from 0 to 1 interpreted as percentages. The higher the value is, the better the model is.",
    func=lambda y, pred, **kwargs: r2_score(y, pred, **kwargs),
    plot_funcs=[(display_single_value, dict(width=900.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    purpose=Purpose.objective,
)
""" ~ """

residuals = Metric[List[float]](
    name="Residuals",
    key="residuals",
    category=MetricCategories.residual,
    func=lambda y, pred, **kwargs: y - pred,
    plot_funcs=[
        (display_acf_plot, dict(width=1500.0)),
        (display_density_plot, dict(width=1500.0)),
        (display_time_series, dict(width=1500.0)),
    ],
    supports_rolling=False,
    purpose=Purpose.diagram,
)
""" ~ """

residuals_mean = Metric[float](
    name="Mean of the Residuals",
    key="residuals_mean",
    category=MetricCategories.residual,
    func=lambda res, *args, **kwargs: res.mean(),
    plot_funcs=[(display_single_value, dict(width=900.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    purpose=Purpose.diagram,
)
""" ~ """

residuals_std = Metric[float](
    name="Standard Deviation of the Residuals",
    key="residuals_std",
    category=MetricCategories.residual,
    func=lambda res, **kwargs: res.std(),
    plot_funcs=[(display_single_value, dict(width=900.0))],
    plot_funcs_rolling=(display_time_series, dict(width=1500.0)),
    purpose=Purpose.diagram,
)
""" ~ """

# ljung_box_statistics = Metric[pd.DataFrame](
#     name="Ljung Box Statistics",
#     key="ljung_box_statistics",
#     category=MetricCategories.residual,
#     func=lambda y, pred, **kwargs: ljung_box(y, pred, **kwargs),
#     info="If p is larger than our significance level then we cannot dismiss the null-hypothesis that the residuals are a random walk.",
#     restrict_to_sample=SampleTypes.insample,
#     comp_complexity=ComputationalComplexity.high,
#     calculation=Calculation.single,
#     purpose=Purpose.diagram,
# )
""" ~ """

residual_group = Group[pd.Series](
    name="residual_group",
    key="residual_group",
    metrics=[residuals_mean, residuals_std],
    preprocess_func=lambda y, pred, prob, **kwargs: y - pred,
    purpose=Purpose.group,
)
""" ~ """

all_regression_metrics = [
    mae,
    mape,
    smape,
    mse,
    rmse,
    # rmsle,
    r_two,
    residuals,
    residual_group,
    # ljung_box_statistics,
]
""" ~ """

minimal_regression_metrics = [
    mae,
    mape,
    mse,
    rmse,
    rmsle,
    r_two,
]
""" ~ """

low_computation_regression_metrics = [
    metric
    for metric in all_regression_metrics
    if metric.comp_complexity is not ComputationalComplexity.high
]
""" ~ """


class RegressionRegistry:
    def __init__(self) -> None:
        self.mae = mae
        self.mape = mape
        self.smape = smape
        self.mse = mse
        self.rmse = rmse
        self.rmsle = rmsle
        self.r_two = r_two
        self.residuals = residuals
        self.residuals_mean = residuals_mean
        self.residuals_std = residuals_std
        self.residual_group = residual_group

        self.all_regression_metrics = deepcopy(all_regression_metrics)
        self.minimal_regression_metrics = deepcopy(minimal_regression_metrics)
        self.low_computation_regression_metrics = deepcopy(
            low_computation_regression_metrics
        )



# from . import library as library
from typing import List

from .compare import compare
from .library_.registry import library
from .metric import Metric
from .score import score
from .scorecard import ScoreCard
from .type import Calculation, DatasetType, MetricCategories, SampleTypes



from typing import List

from ..metric import Metric
from ..type import DatasetType
from .benchmarking_models import ModelRegistry
from .default_metrics_classification import ClassificationRegistry
from .default_metrics_regression import RegressionRegistry


class library:
    ClassificationRegistry = ClassificationRegistry
    RegressionRegistry = RegressionRegistry
    ModelRegistry = ModelRegistry


def get_default_metrics_for_dataset_type(type: DatasetType) -> List[Metric]:
    if type == DatasetType.classification_binary_balanced:
        return ClassificationRegistry().binary_classification_balanced_metrics
    elif type == DatasetType.classification_binary_imbalanced:
        return ClassificationRegistry().binary_classification_imbalanced_metrics
    elif type == DatasetType.classification_multiclass:
        return ClassificationRegistry().multiclass_classification_metrics
    elif type == DatasetType.regression:
        return RegressionRegistry().all_regression_metrics
    else:
        raise ValueError(f"Unknown dataset type {type}")
